unit dlretrieve;

{$mode objfpc}

interface

uses
  Classes, SysUtils, FileUtil, SynMemo, TATools, TASources, TAGraph,
  TAFuncSeries, TASeries, LResources, Forms, Controls, Graphics, Dialogs,
  StdCtrls, Buttons, LCLType, ExtCtrls, ComCtrls, EditBtn
 , UMathInterpolation
 , StrUtils
 , plotter //Plotter view
 //, TADrawUtils
  ;

type

  { TDLRetrieveForm }

  TDLRetrieveForm = class(TForm)
    Block: TButton;
    PlotterButton: TButton;
    CalculatingProgressBar: TProgressBar;
    CalculatingText: TStaticText;
    Chart1: TChart;
    Chart1ColorMapSeries1: TColorMapSeries;
    Chart1LineSeries1: TLineSeries;
    Chart1LineSeries2: TLineSeries;
    Chart2: TChart;
    Chart2ColorMapSeries1: TColorMapSeries;
    Chart2LineSeries1: TLineSeries;
    CursorValue: TStaticText;
    CursorValueGroup: TGroupBox;
    DecorationsGroup: TGroupBox;
    DLRetConvRawButton: TButton;
    DLRetrieveRawButton: TButton;
    ExportButton: TButton;
    HourGlassTimer: TIdleTimer;
    FileDirectoryLabel: TLabel;
    LeftSideLabel: TLabel;
    LegendmaxEntry: TEdit;
    LegendMinEntry: TEdit;
    LogsDirectoryDisplay: TDirectoryEdit;
    LogsDirStatusLabel: TLabel;
    MarkPointsCheckBox: TCheckBox;
    MaxRecordsLabel: TLabel;
    MaxValueLabel: TLabel;
    MinValueLabel: TLabel;
    NorthLabel: TLabel;
    OrientationSelect: TRadioGroup;
    PageControl1: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    PlotFileTitle: TStaticText;
    PlotSettingsGroup: TGroupBox;
    RangeGroup: TGroupBox;
    ResetToLogsDirectoryButton: TBitBtn;
    RightSideLabel: TLabel;
    ScrollBox1: TScrollBox;
    ShowDotsCheckBox: TCheckBox;
    ShowLinesCheckBox: TCheckBox;
    ShowPlotDataButton: TButton;
    SouthLabel: TLabel;
    SynMemo2: TSynMemo;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    ToLabel: TLabel;
    RangeEnd: TEdit;
    RangeStart: TEdit;
    RetRangeButton: TButton;
    ChartToolset1: TChartToolset;
    ChartToolset1PanDragTool1: TPanDragTool;
    ChartToolset1ZoomDragTool1: TZoomDragTool;
    DLCancelRetrieveButton: TButton;
    DLGHeaderButton: TButton;
    DLRetrieveAllButton: TButton;
    ListChartSource1: TListChartSource;
    ListChartSource2: TListChartSource;
    OpenDialog2: TOpenDialog;
    OpenAnotherFileButton: TBitBtn;
    RecentFileEdit: TLabeledEdit;
    OpenDialog1: TOpenDialog;
    OpenRecentFileButton: TBitBtn;
    SynMemo1: TSynMemo;
    procedure BlockClick(Sender: TObject);
    procedure PlotterButtonClick(Sender: TObject);
    procedure DLRetConvRawButtonClick(Sender: TObject);
    procedure DLRetrieveRawButtonClick(Sender: TObject);
    procedure HourGlassTimerTimer(Sender: TObject);
    procedure LogsDirectoryDisplayChange(Sender: TObject);
    procedure OrientationSelectClick(Sender: TObject);
    procedure ResetToLogsDirectoryButtonClick(Sender: TObject);
    procedure RetRangeButtonClick(Sender: TObject);
    procedure ShowDotsCheckBoxChange(Sender: TObject);
    procedure ShowPlotDataButtonClick(Sender: TObject);
    procedure Chart1ColorMapSeries1Calculate(const AX, AY: Double; out
      AZ: Double);
    procedure Chart1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer
      );
    procedure Chart2ColorMapSeries1Calculate(const AX, AY: Double; out
      AZ: Double);
    procedure MarkPointsCheckBoxChange(Sender: TObject);
    procedure DLCancelRetrieveButtonClick(Sender: TObject);
    procedure DLGHeaderButtonClick(Sender: TObject);
    procedure DLRetrieveAllButtonClick(Sender: TObject);
    procedure LegendMinEntryChange(Sender: TObject);
    procedure LegendmaxEntryChange(Sender: TObject);
    procedure ExportButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure OpenAnotherFileButtonClick(Sender: TObject);
    procedure OpenDirectoryButtonClick(Sender: TObject);
    procedure OpenRecentFileButtonClick(Sender: TObject);
    procedure ShowLinesCheckBoxChange(Sender: TObject);
  private
    { private declarations }
    procedure HourGlass();
    procedure DotsLinesUpdate();
  public
    { public declarations }
    procedure DLRetrieveRange(StartRecord:LongInt=1;EndRecord:LongInt=-1);
    procedure UpdateLegend();
end;

var
  DLRetrieveForm: TDLRetrieveForm;

  //FormContourmap: TFormContourmap;
  Int3D: TInterpolation3D;

  VectorPlotOverride: Boolean = False;

  LegendMin: Double = 10.0;
  LegendMax: Double = 22.0;

  HourGlassTimeout: Integer = 1000; //Hourglass shows while below a certain value

implementation

uses
  Unit1, math,appsettings
  ,header_utils,dateutils
  ,dlheader //contains the DLHeader code
  ,dlerase
  ,vector
  ;

{ TDLRetrieveForm }

procedure TDLRetrieveForm.DLGHeaderButtonClick(Sender: TObject);
begin
  DLHeaderForm.ShowModal;
end;

procedure TDLRetrieveForm.DLCancelRetrieveButtonClick(Sender: TObject);
begin
  DLCancelRetrieve:=True;
end;

procedure TDLRetrieveForm.Chart1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var x1,y1,z1: Double;
begin
  If Int3D <> nil then begin
    If Chart1LineSeries1.IsEmpty then Exit;
    x1 := Chart1.XImageToGraph(X);
    y1 := Chart1.YImageToGraph(Y);
    z1 := Int3D.GetIntData(x1,y1);
    CursorValue.Caption :=
      'X: ' + FloatToStrF(x1,ffFixed,6,3) + #13 +
      'Y: ' + FloatToStrF(y1,ffFixed,6,3) + #13 +
      'MPSAS: ' + FloatToStrF(z1,ffFixed,6,3);
  end;
end;

procedure TDLRetrieveForm.Chart2ColorMapSeries1Calculate(const AX, AY: Double;
  out AZ: Double);
begin
  //Show hourglass cursor while changing display.
  HourGlass();
  AZ:=AY;
end;

procedure TDLRetrieveForm.MarkPointsCheckBoxChange(Sender: TObject);
begin
  Chart1LineSeries1.Marks.Visible := MarkPointsCheckBox.Checked;
end;

procedure TDLRetrieveForm.ShowPlotDataButtonClick(Sender: TObject);
var i: Integer;
    SL: TStringList;

    Darkness, Altitude, Azimuth, xValue, yValue: Double;
    pieces: TStringList; //delimited result from generic read
    FirstLine:Integer;
    UdmFeatureVersion:Integer;
    ExpectedFields:Integer;

begin

    Chart1.Visible:=False;
    Chart2.Visible:=False;

    DotsLinesUpdate();

    SL := TStringList.Create;

    pieces := TStringList.Create;
    pieces.Delimiter := '.'; //Initially for determining UDM version
    pieces.StrictDelimiter := True; //Do not parse spaces

    if OpenDialog1.Execute then begin
      //Show hourglass cursor while reading data in.
      HourGlass();

      SL.LoadFromFile(OpenDialog1.FileName); // Quick-and-Dirty data import

      If Int3D <> nil then Int3D.Free;
      Int3D := TInterpolation3D.Create;
      Int3D.ClearData;

      Chart1LineSeries1.BeginUpdate;
      Chart1LineSeries1.Clear;
      Int3D.InterpolationMode := IM_DELAUNAY;


      PlotFileTitle.Caption:=ExtractFileName(OpenDialog1.FileName);

      //Find out which line is the begining
      For i:=0 to SL.Count-1 do begin

        //Show hourglass cursor while reading data in.
        HourGlass();

        //Find UDM version that file was originally stored with
        if AnsiStartsStr('# UDM version:',SL.Strings[i]) then begin
          pieces.DelimitedText := SL.Strings[i];
          UdmFeatureVersion:=StrToIntDef(pieces[3],0);
          StatusMessage('UdmFeatureVersion= '+IntToStr(UdmFeatureVersion));
        end;
        if AnsiContainsStr('# END OF HEADER',SL.Strings[i]) then begin
          break;
        end;
      end;

      FirstLine:=i+1;

      pieces.Delimiter := ';';
      pieces.StrictDelimiter := False; //Parse spaces also

      if UdmFeatureVersion<54 then
          ExpectedFields:=14
      else
          ExpectedFields:=15;

      For i:=FirstLine+1 {without header lines} to SL.Count-1 do begin

        //Show hourglass cursor while reading data in.
        HourGlass();

        pieces.DelimitedText := SL.Strings[i];
        if pieces.Count=ExpectedFields then begin
          Darkness:=StrToFloatDef(pieces[4],0,FPointSeparator);
          Altitude:=StrToFloatDef(pieces[11],0,FPointSeparator);
          if UdmFeatureVersion<54 then begin
              Azimuth:=StrToFloatDef(pieces[12],0,FPointSeparator);
          end
          else
              Azimuth:=StrToFloatDef(pieces[13],0,FPointSeparator);
          xValue:=cos(degtorad(Azimuth+90))*(1.0-abs(Altitude/90.0));
          yValue:=sin(degtorad(Azimuth+90))*(1.0-abs(Altitude/90.0));
          Int3D.AddData(
            xValue,
            yValue,
            Darkness
            );
          Chart1LineSeries1.AddXY(xValue,yValue, FloatToStr(Altitude)+' '+FloatToStr(Azimuth)); // point number as Mark-Label
        end
        else begin
          StatusMessage('expected '+IntToStr(ExpectedFields)+' fields, got '+IntToStr(pieces.Count)+' : '+SL.Strings[i]);
        end;
      end;

    Chart1LineSeries1.EndUpdate;
    Chart1LineSeries2.BeginUpdate;
    Chart1LineSeries2.Clear;

    For i:=0 to High(Int3D.Triangles) do begin

      //Show hourglass cursor while processing data.
      HourGlass();

      Chart1LineSeries2.AddXY(Int3D.Data[Int3D.Triangles[i].a,0],Int3D.Data[Int3D.Triangles[i].a,1]);
      Chart1LineSeries2.AddXY(Int3D.Data[Int3D.Triangles[i].b,0],Int3D.Data[Int3D.Triangles[i].b,1]);
      Chart1LineSeries2.AddXY(Int3D.Data[Int3D.Triangles[i].c,0],Int3D.Data[Int3D.Triangles[i].c,1]);
      Chart1LineSeries2.AddXY(Int3D.Data[Int3D.Triangles[i].a,0],Int3D.Data[Int3D.Triangles[i].a,1]);
      Chart1LineSeries2.AddXY(NaN,NaN); // manual breakpoint (TAChart backward compatible) or AddNull if possible
    end;
    Chart1LineSeries2.EndUpdate;

    //MPSAS fixed values, otherwise range is automatically changed
    Int3D.zmin:=LegendMin;
    Int3D.zmax:=LegendMax;

    LegendMinEntry.Text := FloatToStr(Int3D.zmin); // start values for manual scaling
    LegendmaxEntry.Text := FloatToStr(Int3D.zmax);

    SL.Free;

  end;

  Chart1.Visible:=True;
  Chart2.Visible:=True;

end;

procedure TDLRetrieveForm.Chart1ColorMapSeries1Calculate(const AX, AY: Double;
  out AZ: Double);
begin
  If Int3D <> nil then begin
    HourGlass(); //Make cursor become hourglass
    AZ := Int3D.GetNomalizedIntData(AX,AY);
  end;
end;

procedure TDLRetrieveForm.RetRangeButtonClick(Sender: TObject);
begin
  DLRetrieveRange(StrToIntDef(RangeStart.Text,1),StrToIntDef(RangeEnd.Text,-1));
end;

procedure TDLRetrieveForm.OrientationSelectClick(Sender: TObject);
begin
  HourGlass();
  if OrientationSelect.ItemIndex=0 then begin
    RightSideLabel.Caption:='W';
    LeftSideLabel.Caption:='E';
    Chart1.AxisList[1].Inverted:=False;
  end
  else begin
    RightSideLabel.Caption:='E';
    LeftSideLabel.Caption:='W';
    Chart1.AxisList[1].Inverted:=True;
  end;

end;

procedure TDLRetrieveForm.ResetToLogsDirectoryButtonClick(Sender: TObject);
begin
  {Reset}
  appsettings.LogsDirectoryReset();
  {Update display}
  LogsDirectoryDisplay.Directory:= RemoveMultiSlash(appsettings.LogsDirectory);
  {Save setting}
  vConfigurations.WriteString('Directories','LogsDirectory',appsettings.LogsDirectory);
end;

procedure TDLRetrieveForm.HourGlassTimerTimer(Sender: TObject);
const
  HourGlassTime = 4;
begin
  inc(HourGlassTimeout);
  if (HourGlassTimeout > HourGlassTime) then begin
    CalculatingText.Visible:=False;
    CalculatingProgressBar.Visible:=False;
    HourGlassTimeout:= HourGlassTime;
  end
  else begin
    CalculatingText.Visible:=True;
    CalculatingProgressBar.Visible:=True;
  end;

  Application.ProcessMessages;

end;

procedure TDLRetrieveForm.LogsDirectoryDisplayChange(Sender: TObject);
begin
  {Allow changes to setting}
  appsettings.LogsDirectory:=LogsDirectoryDisplay.Directory;
  {Check validity}
  if DirectoryExists(appsettings.LogsDirectory) then begin
    {Remove warning}
    LogsDirStatusLabel.Caption:='';
    {Save setting}
    vConfigurations.WriteString('Directories','LogsDirectory',appsettings.LogsDirectory);
  end
  else
    {Show warning}
    LogsDirStatusLabel.Caption:='Directory does not exist!';
end;

procedure TDLRetrieveForm.DLRetrieveRawButtonClick(Sender: TObject);
var
    result: String;
    subfix: AnsiString;
    pieces: TStringList;
    TriggerModeNumber: Integer;
    StartTime: TDateTime;
    PacketLength,i: Integer;
    NumberOfPackets,j: LongInt;
    ResultChar: Char;
    ReadingLine: Boolean = True;
    ReadingPackets: Boolean = True;
    OutString: String ='';
begin

    StartTime:=Now;

    SynMemo1.Lines.Clear;

    StatusMessage('DL Retrieve Raw initiated.');

    pieces := TStringList.Create;
    pieces.Delimiter := ',';

    SynMemo2.Clear;
    RecentFileEdit.Text:='';

    //Warn if selected model is not capable of datalogging, this can happen if
    // this window was started from the main menu instead of from the datalogging
    // tab.
    if not((SelectedModel=model_DL) or (SelectedModel=model_V)) then begin
       MessageDlg('Select a datalogging meter',mtWarning, [mbOK],0);
       exit;
    end;

    //Check firmware version before proceeding
    if (StrToInt(SelectedFeature)<45) then begin
       MessageDlg('Upgrade the firmware to feature 45 or greater',mtWarning, [mbOK],0);
       exit;
    end;



    //Ensure that continuous logging mode has not been selected because it
    // interferes with data retrieval on firmware less than feature 30.
    result:=sendget('Lmx');
    TriggerModeNumber:=StrToInt(AnsiMidStr(result,4,1));

    //Warn user that retrieve cannot be done with certain firmwares
    // and certain trigger modes.
    if ( (StrToIntDef(SelectedFeature,0)<30) and ((TriggerModeNumber=1) or (TriggerModeNumber=2))) then begin
      Application.MessageBox('Cannot retrieve database while in continuous trigger mode.'+ sLineBreak+
      'Select trigger mode = Off,'+ sLineBreak+
      ' or one of the "Every x on the hour" modes.', 'Busy logging, cannot retrieve!', MB_ICONEXCLAMATION)  ;
      exit;
    end;

     if ((Length(DLHeaderForm.TZRegionBox.Text)>0) and (Length(DLHeaderForm.TZLocationBox.Text)>0)) then
       begin
       DLCancelRetrieve:=False;
         //Get unit time for comparison
         try
            if (SelectedModel=model_V) then
              WriteDLHeader('DL-V-Log','DL Retrieve Raw', '.raw') //vector model
            else
              WriteDLHeader('DL-Log','DL Retrieve Raw', '.raw');// Standard DL model

            //Open file for appending records
            AssignFile(DLRecFile, LogFileName);
            Reset(DLRecFile);
            Append(DLRecFile); //Open file for appending
            SetTextLineEnding(DLRecFile, #13#10);

            OpenComm();
            ser.SendString('L8x');
            //Pull in first line which contains packet length, and number of packets.
            While (ReadingLine) do begin
              ResultChar:=chr(ser.RecvByte(2000));
              if ResultChar=#10 then
                ReadingLine:=False;
              OutString:=OutString+ResultChar;
            end;
            pieces.DelimitedText:=OutString;
            PacketLength:=StrToInt(pieces.Strings[1]);
            NumberOfPackets:=StrToInt64(pieces.Strings[2]);
            Writeln(DLRecFile,Format('# PacketLength: %d',[PacketLength]));
            Writeln(DLRecFile,Format('# NumberOfPackets: %d',[NumberOfPackets]));
            Writeln(DLRecFile,'# END OF HEADER');


            { Write records to rawfile }
            for j:= 1 to NumberOfPackets do begin
              SynMemo1.Lines.Clear;
              SynMemo1.Lines.Append(Format('Retrieved record: %d / %d',[j, NumberOfPackets]));
              Application.ProcessMessages;
              for i:= 1 to PacketLength do begin
                ResultChar:=chr(ser.RecvByte(1000));
                If CompareStr(ser.LastErrorDesc,'OK')<>0 then begin
                   StatusMessage('Error: '+ser.LastErrorDesc);
                   ReadingPackets:=False;
                   Break;
                end else begin
                  Write(DLRecFile,ResultChar);
                end;
                if not(ReadingPackets) then Break;
              end;
            end;

            //Read EOF
            ReadingLine:=True;
            OutString:='';
            While (ReadingLine) do begin
              ResultChar:=chr(ser.RecvByte(1000));
              if ResultChar=#10 then
                ReadingLine:=False;
              write(DLRecFile,ResultChar);
            end;

            CloseComm;

            SynMemo1.Lines.Clear;
             if DLCancelRetrieve then
                SynMemo1.Lines.Append('Partially retrieved ' +
                  IntToStr(j)+' records: ' +
                  '1 to ' + IntToStr(i-1) +'  written to:')
             else
                 SynMemo1.Lines.Append('Retrieved ' +
                 IntToStr(j)+' records: ' +
                 '1 to ' + IntToStr(NumberOfPackets) +' written to:');

             RecentFileEdit.Text:=Format('%s',[LogFileName]);
             SynMemo1.Lines.Append(Format('%s',[LogFileName]));
             DLCancelRetrieve:=False;

             Flush(DLRecFile);
             CloseFile(DLRecFile);
         except
               StatusMessage('ERROR! IORESULT: ' + IntToStr(IOResult));
         end;
       end
     else
         ShowMessage('Enter Time zone information into Header first. '+ sLineBreak+
         'Do this by pressing the Header button, then selecting your timezone.');

     SynMemo1.Lines.Append(Format('Time to retieve raw data: %.3f seconds.',[MilliSecondsBetween(Now,StartTime)/1000.]));
end;


procedure TDLRetrieveForm.DLRetConvRawButtonClick(Sender: TObject);
var
    subfix: AnsiString;
    Temperature, Voltage, Darkness : Float;
    pieces: TStringList;
    StartTime: TDateTime;
    LogTime : TDateTime;
    NumberOfPackets: LongInt;
    PacketLength: Integer;
    j: LongInt=0;
    ResultChar: Char;
    InFile,OutFile: TextFile;
    OutFileString: String;
    Str: String;
    SS,MN,HH,DY,DT,MO,YY: String;
    UTCDateString: String;
    WriteAllowable: Boolean = True; //Allow output file to be written or not.

function CharToHex():String;
begin
  Read(InFile,ResultChar);
  CharToHex:=IntToHex(Ord(ResultChar),2);
end;

function CharToHexLimit():String;
begin
  Read(InFile,ResultChar);
  CharToHexLimit:=IntToStr(StrToIntDef(IntToHex(Ord(ResultChar),2),0));
end;

function CharsToLongInt():LongInt;
//32 bit read in. Sign is maintained becauase LongInt is 32 bit.
begin
  Read(InFile,ResultChar);
  CharsToLongInt:=Ord(ResultChar);
  Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(ResultChar)*256;
  Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(ResultChar)*256**2;
  Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(ResultChar)*256**3;
end;

function CharsToSmallInt():SmallInt;
//16 bit read in. Sign is maintained becauase SmallInt is 16 bit.
begin
  Read(InFile,ResultChar);
  CharsToSmallInt:=Ord(ResultChar);
  Read(InFile,ResultChar);
  CharsToSmallInt:=CharsToSmallInt+Ord(ResultChar)*256;
end;

function CharToByte():Byte;
//8 bit read in. Unsigned.
begin
  Read(InFile,ResultChar);
  CharToByte:=byte(ResultChar);
end;

begin

  SynMemo1.Lines.Clear;

    StatusMessage('DL Raw to .dat conversion initiated.');

    pieces := TStringList.Create;

    OpenDialog1.Filter:='raw log files|*.raw|All files|*.*';
    OpenDialog1.InitialDir := appsettings.LogsDirectory;

     if OpenDialog1.Execute then begin
         //Start reading file.

         AssignFile(InFile, OpenDialog1.Filename);
         OutFileString:=ChangeFileExt(OpenDialog1.Filename,'.dat');
         AssignFile(OutFile, OutFileString);
         if FileExists(OutFileString) then begin
             if (MessageDlg('Overwrite existing file?','Do you want to overwrite the existing file?',mtConfirmation,[mbOK,mbCancel],0) = mrOK) then
                 WriteAllowable:=True
             else
               WriteAllowable:=False;
           end;

         if WriteAllowable then begin
         StartTime:=Now;

         {$I+} // Errors will lead to an EInOutError exception (default)
         try
           Reset(InFile);

           Rewrite(OutFile); //Open file for writing

           repeat
             // Read one line at a time from the file.
             Readln(InFile, Str);
             pieces.Delimiter := ' ';
             pieces.StrictDelimiter := False; //Parse spaces.

             // Get location data from header.
             if AnsiStartsStr('# UDM setting:',Str) then
                 Str:='# UDM setting: DL .raw converted to .dat';
             if AnsiStartsStr('# PacketLength:',Str) then begin
                pieces.DelimitedText:=Str;
                PacketLength:=StrToInt(pieces.Strings[2]);
             end;
             if AnsiStartsStr('# NumberOfPackets:',Str) then begin
                pieces.DelimitedText:=Str;
                NumberOfPackets:=StrToInt(pieces.Strings[2]);;
             end;

             If Str<>'' then
                WriteLn(OutFile, Str);

           until(AnsiStartsStr('# END OF HEADER',Str)); // End Of header.

           for j:=1 to NumberOfPackets do begin
           //Pull in binary
           //
           Read(InFile,ResultChar); //Position 0, Indicates if EEPROM value was written, throw away.
           SS:=CharToHex; //Position 1, Seconds.
           MN:=CharToHex; //Position 2, Minutes.
           HH:=CharToHex; //Position 3, Hours.
           DY:=CharToHex; //Position 4, Day.
           DT:=CharToHex; //Position 5, Date.
           MO:=CharToHex; //Position 6, Month.
           YY:=CharToHex; //Position 7, Year.
           UTCDateString:= YY+'-'+MO+'-'+DT+HH+':'+MN+':'+SS;

           try
             //Convert datestring to date for local time calculation
             LogTime:=ScanDateTime('yy-mm-ddhh:nn:ss',UTCDateString);
           except
                 LogTime:=ScanDateTime('yy-mm-ddhh:nn:ss','00-00-0000:00:00');
           end;

           //{ Pull in values }
           //;	LRECDATA+8-11	Rdg   raw 32 bit binary value from CDMAGS
           Darkness:=StrToFloatDef(((Format('%.2f',[CharsToLongInt / 6553600.0]))),0);

           //;	LRECDATA+12,13	Temp. raw 16 bit binary value
           Temperature:=(((CharsToSmallInt * 33000.0) / 1024.0) -5000.0) / 100.0;

           //;	LRECDATA+14	Batt. raw 8 bit binary value
           Voltage:=(2.048 + (3.3 * CharToByte)/256.0);

           Read(InFile,ResultChar); //Position 15, unused spare, throw away.

           { Write record to logfile }
           Writeln(OutFile,
              FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',LogTime) + //Date UTC
              FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',dlheader.ptz.GMTToLocalTime(LogTime,DLHeaderForm.TZLocationBox.Text,subfix)) + //Date Local (calculated)
              FormatFloat('##0.0',Temperature,FPointSeparator) + ';' + //Temperature
              FormatFloat('0.00',Voltage,FPointSeparator) + ';' + //Voltage
              FormatFloat('#0.00',Darkness,FPointSeparator) //mpsas value
           );
           end;

           CloseFile(InFile);
         except
           on E: EInOutError do begin
            MessageDlg('Error', 'File handling error occurred. Details: '+E.ClassName+'/'+E.Message, mtError, [mbOK],0);
           end;
         end;
         Flush(OutFile);
         CloseFile(OutFile);

         //Display file information
         SynMemo1.Clear;
         SynMemo1.Lines.Append('Retrieved ' +
               IntToStr(j)+' records: ' +
               '1 to ' + IntToStr(j) +' written to:');
         SynMemo1.Lines.Append(Format('%s',[OutFileString]));

         //Display conversion time
         SynMemo1.Lines.Append(Format('Time to convert raw data: %.3f seconds.',[MilliSecondsBetween(Now,StartTime)/1000.]));

         end;//End of WriteAllowable check.

       end;


     //read in selected file
     //copy header lines mostly verbatim

     //get packetsize and packetnumber

     //convert date, temp, mpsas, voltage


end;

procedure TDLRetrieveForm.BlockClick(Sender: TObject);
var
    ErrorReading:Boolean = False;
    j: LongInt=0;
    LogTime : TDateTime;
    NumberOfPackets: LongInt;
    OutString: String ='';
    PacketLength: Integer;
    pieces: TStringList;
    ReadingLine: Boolean = True;
    result: String;
    ResultChar: Char;
    FirstChar:Byte;
    SS,MN,HH,DY,DT,MO,YY: String;
    StartTime: TDateTime;
    subfix: AnsiString;
    Temperature, Voltage, Darkness : Float;
    Vibration : SmallInt;
    Altitude:Double;
    Arx,Ary,Arz,Mrx,Mry,Mrz: SmallInt; //Raw accellereometer and magnetometer data from meter
    TriggerModeNumber: Integer;
    UTCDateString: String;

function CharToHex():String;
begin
  //Read(InFile,ResultChar);
  CharToHex:=IntToHex(Ord(chr(ser.RecvByte(1000))),2);
end;

function CharToHexLimit():String;
begin
  //Read(InFile,ResultChar);
  CharToHexLimit:=IntToStr(StrToIntDef(IntToHex(Ord(chr(ser.RecvByte(1000))),2),0));
end;

function CharsToLongInt():LongInt;
//32 bit read in. Sign is maintained becauase LongInt is 32 bit.
begin
  //Read(InFile,ResultChar);
  CharsToLongInt:=Ord(chr(ser.RecvByte(1000)));
  //Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(chr(ser.RecvByte(1000)))*256;
  //Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(chr(ser.RecvByte(1000)))*256**2;
  //Read(InFile,ResultChar);
  CharsToLongInt:=CharsToLongInt+Ord(chr(ser.RecvByte(1000)))*256**3;
end;

function CharsToSmallInt():SmallInt;
//16 bit read in. Sign is maintained because SmallInt is 16 bit (-32768 to 32767).
begin
  //Read(InFile,ResultChar);
  CharsToSmallInt:=Ord(chr(ser.RecvByte(1000)));
  //Read(InFile,ResultChar);
  CharsToSmallInt:=CharsToSmallInt+Ord(chr(ser.RecvByte(1000)))*256;
end;

function CharToByte():Byte;
//8 bit read in. Unsigned.
begin
  //Read(InFile,ResultChar);
  CharToByte:=byte(chr(ser.RecvByte(1000)));
end;
begin

    StartTime:=Now;

    SynMemo1.Lines.Clear;

    StatusMessage('DL binary retrieve initiated.');

    pieces := TStringList.Create;
    pieces.Delimiter := ',';

    SynMemo2.Clear;
    RecentFileEdit.Text:='';

    //Warn if selected model is not capable of datalogging, this can happen if
    // this window was started from the main menu instead of from the datalogging
    // tab.
    if not((SelectedModel=model_DL) or (SelectedModel=model_V)) then begin
       MessageDlg('Select a datalogging meter',mtWarning, [mbOK],0);
       exit;
    end;

    //Check firmware version before proceeding
    if (StrToInt(SelectedFeature)<47) then begin
       MessageDlg('Upgrade the firmware to feature 45 or greater',mtWarning, [mbOK],0);
       exit;
    end;

    //Ensure that continuous logging mode has not been selected because it
    // interferes with data retrieval on firmware less than feature 30.
    result:=sendget('Lmx');
    TriggerModeNumber:=StrToInt(AnsiMidStr(result,4,1));

    //Warn user that retrieve cannot be done with certain firmwares
    // and certain trigger modes.
    if ( (StrToIntDef(SelectedFeature,0)<30) and ((TriggerModeNumber=1) or (TriggerModeNumber=2))) then begin
      Application.MessageBox('Cannot retrieve database while in continuous trigger mode.'+ sLineBreak+
      'Select trigger mode = Off,'+ sLineBreak+
      ' or one of the "Every x on the hour" modes.', 'Busy logging, cannot retrieve!', MB_ICONEXCLAMATION)  ;
      exit;
    end;

    // Check that header timezone information has been entered
    if ((Length(DLHeaderForm.TZRegionBox.Text)>0) and (Length(DLHeaderForm.TZLocationBox.Text)>0)) then begin
       DLCancelRetrieve:=False;
         //Get unit time for comparison
         try
            if (SelectedModel=model_V) then
              WriteDLHeader('DL-V-Log','DL-V binary retrieve') //Vector model
            else
              WriteDLHeader('DL-Log','DL binary retrieve');// Standard DL model

            //Open file for appending records
            AssignFile(DLRecFile, LogFileName);
            Reset(DLRecFile);
            Append(DLRecFile); //Open file for appending
            SetTextLineEnding(DLRecFile, #13#10);

            OpenComm();
            ser.SendString('L8x');
            //Pull in first line which contains packet length, and number of packets.
            While (ReadingLine) do begin
              ResultChar:=chr(ser.RecvByte(2000));
              if ResultChar=#10 then
                ReadingLine:=False;
              OutString:=OutString+ResultChar;
            end;

            pieces.DelimitedText:=OutString;
            PacketLength:=StrToInt(pieces.Strings[1]);
            NumberOfPackets:=StrToInt64(pieces.Strings[2]);
            StatusMessage(Format('%d Packets %d bytes long.',[NumberOfPackets, PacketLength]));

            {Read records from meter, and Write records to file.}
            for j:= 1 to NumberOfPackets do begin
              //Send request for second data packet and so on before they are sent.
              ser.SendString('x');

              SynMemo1.Lines.Clear;
              SynMemo1.Lines.Append(Format('Retrieved record: %d / %d',[j, NumberOfPackets]));
              Application.ProcessMessages;

              //Pull in DL binary data
              FirstChar:=CharToByte; //Position 0, Indicates record details.
              SS:=CharToHex; //Position 1, Seconds.
              MN:=CharToHex; //Position 2, Minutes.
              HH:=CharToHex; //Position 3, Hours.
              DY:=CharToHex; //Position 4, Day.
              DT:=CharToHex; //Position 5, Date.
              MO:=CharToHex; //Position 6, Month.
              YY:=CharToHex; //Position 7, Year.
              UTCDateString:= YY+'-'+MO+'-'+DT+HH+':'+MN+':'+SS;

              try
                //Convert datestring to date for local time calculation
                LogTime:=ScanDateTime('yy-mm-ddhh:nn:ss',UTCDateString);
              except
                    LogTime:=ScanDateTime('yy-mm-ddhh:nn:ss','00-00-0000:00:00');
              end;

              //;	LRECDATA+8-11	Rdg   raw 32 bit binary value from CDMAGS
              Darkness:=StrToFloatDef(((Format('%.2f',[CharsToLongInt / 6553600.0]))),0);

              //;	LRECDATA+12,13	Temp. raw 16 bit binary value
              Temperature:=(((CharsToSmallInt * 33000.0) / 1024.0) -5000.0) / 100.0;

              //;	LRECDATA+14	Batt. raw 8 bit binary value
              Voltage:=(2.048 + (3.3 * CharToByte)/256.0);

              ResultChar:=chr(ser.RecvByte(1000)); //Position 15, unused spare, throw away.

              //Read V model data
              if (SelectedModel=model_V) then begin
                ResultChar:=chr(ser.RecvByte(1000)); //Position 16, unused spare, throw away.
                ResultChar:=chr(ser.RecvByte(1000)); //Position 17, unused spare, throw away.

                //; Vibration XYZ max 16 bit binary value         (2) LRECDATA+18 to 19
                Vibration:=CharsToSmallInt();

                Arx:=CharsToSmallInt();//Accel X Rdg raw 16 bit binary value           (2) LRECDATA+20 to 21
                Ary:=CharsToSmallInt();//Accel Y Rdg raw 16 bit binary value           (2) LRECDATA+22 to 23
                Arz:=CharsToSmallInt();//Accel Z Rdg raw 16 bit binary value           (2) LRECDATA+24 to 25
                Mrx:=CharsToSmallInt();//Mag   X Rdg raw 16 bit binary value           (2) LRECDATA+26 to 27
                Mry:=CharsToSmallInt();//Mag   X Rdg raw 16 bit binary value           (2) LRECDATA+28 to 29
                Mrz:=CharsToSmallInt();//Mag   X Rdg raw 16 bit binary value           (2) LRECDATA+30 to 31

                Ax:= Arx * -1.0;
                Ay:= Ary;
                Az:= Arz;
                NormalizeAccel(); //Compute acceleration values (In the future, this may be done inside the PIC)
                //Altitude:=radtodeg(arcsin(-1.0*Ax1));
                Altitude:=ComputeAltitude(Ax1, Ay1, Az1);
                //
                Mx:= Mrx;
                My:= Mry * -1.0;
                Mz:= Mrz * -1.0;
                NormalizeMag();
                ComputeAzimuth();
                Heading:=radtodeg(arctan2(-1*Mz2,Mx2))+180;
              end;

              //check for error in receiving
              If CompareStr(ser.LastErrorDesc,'OK')<>0 then begin
                 StatusMessage('Error receiving data: '+ser.LastErrorDesc +' at record #: '+IntToStr(j));
                 ErrorReading:=True;
                 Break;
              end;

              { Write DL record to logfile }
              Write(DLRecFile,
                 FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',LogTime) + //Date UTC
                 FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',dlheader.ptz.GMTToLocalTime(LogTime,DLHeaderForm.TZLocationBox.Text,subfix)) + //Date Local (calculated)
                 FormatFloat('##0.0',Temperature,FPointSeparator) + ';' + //Temperature
                 FormatFloat('0.00',Voltage,FPointSeparator) + ';' + //Voltage
                 FormatFloat('#0.00',Darkness,FPointSeparator) //mpsas value
              );

              { Write V record to logfile }
              if (SelectedModel=model_V) then begin
                 Write(DLRecFile,Format(';%d;',[Arx]));//Acceleration X
                 Write(DLRecFile,Format('%d;',[Ary]));//Acceleration Y
                 Write(DLRecFile,Format('%d;',[Arz]));//Acceleration Z
                 Write(DLRecFile,Format('%d;',[Mrx]));//Magnetic X
                 Write(DLRecFile,Format('%d;',[Mry]));//Magnetic Y
                 Write(DLRecFile,Format('%d;',[Mrz]));//Magnetic Z
                 Write(DLRecFile,Format('%0.1f;',[Altitude])); //Altitude
                 Write(DLRecFile,Format('%0.1f;',[abs(Altitude-90.0)])); //Zenith
                 Write(DLRecFile,Format('%0.0f;',[Heading])); //Azimuth
                 Write(DLRecFile,Format('%d',[Vibration]));//Vibration count
              end;

              if StrToInt(SelectedFeature)>=49 then begin //1st byte contains record type
                if ((FirstChar and %00010000) =0) then //Mask for initial/subsequent indicator
                 Write(DLRecFile,';0') // Initial record
                else
                 Write(DLRecFile,';1') // Subsequent record
              end;

              Writeln(DLRecFile,''); //EOL

            end;

            if ErrorReading then begin
               Writeln(DLRecFile,'There was an error reading meter: '+ser.LastErrorDesc);
            end else begin
              //Read EOF
              ReadingLine:=True;
              While (ReadingLine) do begin
                ResultChar:=chr(ser.RecvByte(1000));
                if ResultChar=#10 then
                  ReadingLine:=False;
              end;
            end;

            CloseComm;

            SynMemo1.Lines.Clear;
            if ErrorReading then begin
               SynMemo1.Lines.Append('>>>  Error receiving data <<<');
               SynMemo1.Lines.Append('Retrieved only ' +
                 IntToStr(j)+' records of ' +
                 IntToStr(NumberOfPackets) +' total. Written to:');
            end else begin
              if DLCancelRetrieve then
                  SynMemo1.Lines.Append('Partially retrieved ' +
                    IntToStr(j)+' records: ' +
                    '1 to ' + IntToStr(j-1) +'  written to:')
              else
                   SynMemo1.Lines.Append('Retrieved ' +
                   IntToStr(j)+' records: ' +
                   '1 to ' + IntToStr(NumberOfPackets) +' written to:');
            end;

             RecentFileEdit.Text:=Format('%s',[LogFileName]);
             SynMemo1.Lines.Append(Format('%s',[LogFileName]));
             DLCancelRetrieve:=False;

             Flush(DLRecFile);
             CloseFile(DLRecFile);
         except
               StatusMessage('ERROR! IORESULT: ' + IntToStr(IOResult)+' at record #: '+IntToStr(j));
         end;
         SynMemo1.Lines.Append(Format('Operation time: %.3f seconds.',[MilliSecondsBetween(Now,StartTime)/1000.]));
       end
     else
         ShowMessage('Enter Time zone information into Header first. '+ sLineBreak+
         'Do this by pressing the Header button, then selecting your timezone.');


end;

procedure TDLRetrieveForm.PlotterButtonClick(Sender: TObject);
begin
    plotter.PlotterForm.Show;
end;

procedure TDLRetrieveForm.DLRetrieveAllButtonClick(Sender: TObject);
begin
  DLRetrieveRange();
end;

procedure TDLRetrieveForm.DLRetrieveRange(StartRecord:LongInt=1;EndRecord:LongInt=-1);
var
  result: String;
  subfix: AnsiString;
  pieces: TStringList;
  LogTime : TDateTime;
  Temperature, Voltage, Darkness : Float;
  TriggerModeNumber: Integer;
  DesiredPieces: Integer;
  RecordsRetrieved: Longint;
  Altitude:Double;
  StartTime: TDateTime;
  RecordType:Integer;//0=Initial, 1=Subsequent record type.
begin

  StartTime:=Now;

  RecordsRetrieved:= 0;

  DLCancelRetrieveButton.Enabled:=True;
  pieces := TStringList.Create;
  pieces.Delimiter := ',';

  SynMemo2.Clear;
  RecentFileEdit.Text:='';

  //Warn if selected model is not capable of datalogging, this can happen if
  // this window was started from the main menu instead of from the datalogging
  // tab.
  if not((SelectedModel=model_DL) or (SelectedModel=model_V)) then begin
     MessageDlg('Select a datalogging meter',mtWarning, [mbOK],0);
     exit;
  end;

  //Ensure that continuous logging mode has not been selected because it
  // interferes with data retrieval on firmware less than feature 30.
  result:=sendget('Lmx');
  TriggerModeNumber:=StrToInt(AnsiMidStr(result,4,1));

  //Warn user that retrieve cannot be done with certain firmwares
  // and certain trigger modes.
  if ( (StrToIntDef(SelectedFeature,0)<30) and ((TriggerModeNumber=1) or (TriggerModeNumber=2))) then begin
    Application.MessageBox('Cannot retrieve database while in continuous trigger mode.'+ sLineBreak+
    'Select trigger mode = Off,'+ sLineBreak+
    ' or one of the "Every x on the hour" modes.', 'Busy logging, cannot retrieve!', MB_ICONEXCLAMATION)  ;
    exit;
  end;

   if ((Length(DLHeaderForm.TZRegionBox.Text)>0) and (Length(DLHeaderForm.TZLocationBox.Text)>0)) then begin
     DLCancelRetrieve:=False;
       //Get unit time for comparison
       try
          if (SelectedModel=model_V) then
            WriteDLHeader('DL-V-Log','DL Retrieve All') //vector model
          else
            WriteDLHeader('DL-Log','DL Retrieve All');// Standard DL model

          //Set the desired pieces to be seen when getting a record
          // There are two extra pieces than normal since the date record is split up
          //if SelectedModel=model_V then
          //   DesiredPieces:=14    //vector model
          //else
          //   DesiredPieces:=7;    //other DL models

          if StrToInt(SelectedFeature)>=49 then // feature 49 and above have 1st record indicator
             if SelectedModel=model_V then DesiredPieces:=15 else DesiredPieces:=8
          else //older than feature 49 (did not have 1st record indicator)
              if SelectedModel=model_V then DesiredPieces:=14 else DesiredPieces:=7;


          //Define last record to get if not already defined in the procedure call.
          if EndRecord<0 then begin
             StatusMessage('DLRetrieveRange: End record was '+IntToStr(EndRecord)+', set to '+IntToStr(DLEStoredRecords));
             EndRecord:=DLEStoredRecords;
          end;

          //Check start record
          if StartRecord<1 then begin
             StartRecord:=1;
             StatusMessage('DLRetrieveRange: Start record was '+IntToStr(StartRecord)+', set to 1.');
          end;
          if StartRecord>=EndRecord then begin
             StatusMessage('DLRetrieveRange: Start record >= End record.');
           end;

          //Open file for appending records
          AssignFile(DLRecFile, LogFileName);
          Reset(DLRecFile);
          Append(DLRecFile); //Open file for appending
          SetTextLineEnding(DLRecFile, #13#10);

          for DLCurrentRecord:=StartRecord to EndRecord do begin
              if DLCancelRetrieve then begin
                 StatusMessage('DL Retrieve cancelled.');
                 break;
               end;
               pieces.DelimitedText := sendget(Format('L4%010.10dx',[DLCurrentRecord-1]));
               pieces.StrictDelimiter := False; //Parse spaces also because date is split up

               SynMemo1.Lines.Clear;
               if (pieces.Count=DesiredPieces) then begin
                 SynMemo1.Lines.Append(Format('Retrieved record: %d / %d',[DLCurrentRecord, EndRecord]));
                 Application.ProcessMessages;

                 { Pull in unit time for UTC/Local calculations. }
                 try
                    LogTime:=ScanDateTime('yy-mm-ddhh:nn:ss',pieces.Strings[1]+pieces.Strings[3]);
                 except
                    LogTime:=ScanDateTime('yy-mm-dd hh:nn:ss','01-01-01 01:01:01');//Record something anyway even if local date is messed up
                 end;

                 { Pull in values }
                 Temperature:=StrToFloatDef(AnsiLeftStr(pieces.Strings[5],Length(pieces.Strings[5])-1),0,FPointSeparator);
                 Voltage:=(2.048 + (3.3 * StrToFloatDef(pieces.Strings[6],0,FPointSeparator))/256.0);
                 Darkness:=StrToFloatDef(pieces.Strings[4],0,FPointSeparator);

                 { Write record to logfile }
                 Write(DLRecFile,
                    FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',LogTime) + //Date UTC
                    FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',dlheader.ptz.GMTToLocalTime(LogTime,DLHeaderForm.TZLocationBox.Text,subfix)) + //Date Local (calculated)
                    FormatFloat('##0.0',Temperature,FPointSeparator) + ';' + //Temperature
                    FormatFloat('0.00',Voltage,FPointSeparator) + ';' + //Voltage
                    FormatFloat('#0.00',Darkness,FPointSeparator) //mpsas value
                 );

                 if SelectedModel=model_V then begin
                   Write(DLRecFile,Format(';%d;',[StrToIntDef(pieces.Strings[7],0)]));//Acceleration X
                   Write(DLRecFile,Format('%d;',[StrToIntDef(pieces.Strings[8],0)]));//Acceleration Y
                   Write(DLRecFile,Format('%d;',[StrToIntDef(pieces.Strings[9],0)]));//Acceleration Z
                   Write(DLRecFile,Format('%d;',[StrToIntDef(pieces.Strings[10],0)]));//Magnetic X
                   Write(DLRecFile,Format('%d;',[StrToIntDef(pieces.Strings[11],0)]));//Magnetic Y
                   Write(DLRecFile,Format('%d;',[StrToIntDef(pieces.Strings[12],0)]));//Magnetic Z
                   Ax:= -1.0 * StrToFloatDef(pieces.Strings[7],0,FPointSeparator);
                   Ay:=        StrToFloatDef(pieces.Strings[8],0,FPointSeparator);
                   Az:=        StrToFloatDef(pieces.Strings[9],0,FPointSeparator);
                   NormalizeAccel(); //Compute acceleration values (In the future, this may be done inside the PIC)
                   //Altitude:=radtodeg(arcsin(-1.0*Ax1));
                   Altitude:=ComputeAltitude(Ax1, Ay1, Az1);
                   Write(DLRecFile,Format('%0.0f;',[Altitude]));
                   Write(DLRecFile,Format('%0.0f;',[abs(Altitude-90.0)]));
                   //
                   Mx:=        StrToFloatDef(pieces.Strings[10],0,FPointSeparator);
                   My:= -1.0 * StrToFloatDef(pieces.Strings[11],0,FPointSeparator);
                   Mz:= -1.0 * StrToFloatDef(pieces.Strings[12],0,FPointSeparator);
                   NormalizeMag();
                   ComputeAzimuth();
                   Heading:=radtodeg(arctan2(-1*Mz2,Mx2))+180;
                   Write(DLRecFile,Format('%0.0f;',[Heading]));
                   Write(DLRecFile,Format('%d',[StrToIntDef(pieces.Strings[13],0)]));//Vibration count

                   if StrToInt(SelectedFeature)>=49 then //Lasst peice contains record type
                      RecordType:=StrToIntDef(pieces.Strings[14],0);
                 end else
                 if StrToInt(SelectedFeature)>=49 then //Last peice contains record type
                   RecordType:=StrToIntDef(pieces.Strings[7],0);


                 if StrToInt(SelectedFeature)>=49 then begin //1st byte contains record type
                   if (RecordType=0) then //Check for initial/subsequent indicator
                    Write(DLRecFile,';0') // Initial record
                   else
                    Write(DLRecFile,';1') // Subsequent record
                 end;

                 writeln(DLRecFile,'');//Finish writing line to logfile
               end
               else begin
                  StatusMessage(Format('Failed: actual count = %d , desired count = %d ',[pieces.Count, DesiredPieces]));
                  SynMemo1.Lines.Append(Format('Failed: actual count = %d , desired count = %d ',[pieces.Count, DesiredPieces]));
                  //break;
               end;

               Inc(RecordsRetrieved);

          end;
           SynMemo1.Lines.Clear;
           if DLCancelRetrieve then
              SynMemo1.Lines.Append('Partially retrieved ' +
                IntToStr(RecordsRetrieved)+' records: ' +
                IntToStr(StartRecord) +
                ' to ' + IntToStr(DLCurrentRecord-1) +'  written to:')
           else
               SynMemo1.Lines.Append('Retrieved ' +
               IntToStr(RecordsRetrieved)+' records: ' +
               IntToStr(StartRecord) +
               ' to ' + IntToStr(EndRecord) +' written to:');

           RecentFileEdit.Text:=Format('%s',[LogFileName]);
           SynMemo1.Lines.Append(Format('%s',[LogFileName]));
           DLCancelRetrieve:=False;

           Flush(DLRecFile);
           CloseFile(DLRecFile);
       except
             StatusMessage('ERROR! IORESULT: ' + IntToStr(IOResult));
       end;
     end
   else
       ShowMessage('Enter Time zone information into Header first. '+ sLineBreak+
       'Do this by pressing the Header button, then selecting your timezone.');

   DLCancelRetrieveButton.Enabled:=False;
   SynMemo1.Lines.Append(Format('Time to retieve processed data: %.3f seconds.',[MilliSecondsBetween(Now,StartTime)/1000.]));


end;

procedure TDLRetrieveForm.LegendMinEntryChange(Sender: TObject);
var tmp: Double;
begin
  If Int3D <> nil then begin
    If TryStrToFloat(LegendMinEntry.Text,tmp,FPointSeparator) then begin
      LegendMin:=tmp;
      UpdateLegend();
      Int3D.zmin := tmp;
      Chart1.Invalidate;
    end;
  end;
end;

procedure TDLRetrieveForm.LegendmaxEntryChange(Sender: TObject);
var tmp: Double;
begin
  If Int3D <> nil then begin
    If TryStrToFloat(LegendmaxEntry.Text,tmp,FPointSeparator) then begin
      LegendMax:=tmp;
      UpdateLegend();
      Int3D.zmax := tmp;
      Chart1.Invalidate;
    end;
  end;
end;

procedure TDLRetrieveForm.ExportButtonClick(Sender: TObject);
begin
  Chart1.SaveToFile(TJPEGImage, 'MyImage.jpg');
end;

procedure TDLRetrieveForm.FormCreate(Sender: TObject);
begin

  //Initialize the start and end range for retrieving
  RangeStart.Text:='1';
  RangeEnd.Text:= '-1';

  PageControl1.PageIndex:=0; //Always start with displaying the "Text" page

  //Initialize range of colours for plot chart (values 1 to -1)
  With ListChartSource1 do begin
    Clear;
    Add(1,0,'',$000000);    // Black
    Add(0.6,0,'',$640054);  // Dark purple
    Add(0.3,0,'',$FF0000);  // Blue
    Add(0.1,0,'',$FFFF00);  // Turquise
    Add(0.0,0,'',$7FFF00);
    Add(-0.1,0,'',$00FF00); // Gren
    Add(-0.3,0,'',$00FF7F);
    Add(-0.6,0,'',$00FFFF); // Yellow
    Add(-0.9,0,'',$0000FF); // Red
    Add(-1.0,0,'',$FFFFFF); // White
  end;

  UpdateLegend();
end;

procedure TDLRetrieveForm.UpdateLegend();
var
  LegendRange:Double;
begin
  //Initialize range of colours for legend chart (values 22 to 10)
  LegendRange:=LegendMax-LegendMin;
  With ListChartSource2 do begin
    Clear;
    Add(LegendMax,0,'',$000000);  // Black
    Add(LegendMin + LegendRange * 0.80, 0, '', $640054);  // Dark purple
    Add(LegendMin + LegendRange * 0.65, 0, '', $FF0000);  // Blue1
    Add(LegendMin + LegendRange * 0.55, 0, '', $FFFF00);  // cyan1
    Add(LegendMin + LegendRange * 0.50, 0, '', $7FFF00);  // SpringGreen1
    Add(LegendMin + LegendRange * 0.45, 0, '', $00FF00);  // Green1
    Add(LegendMin + LegendRange * 0.35, 0, '', $00FF7F);  // chartreuse1
    Add(LegendMin + LegendRange * 0.20, 0, '', $00FFFF);  // Yellow1
    Add(LegendMin + LegendRange * 0.05, 0, '', $0000FF);  // Red1
    Add(LegendMin,0,'',$FFFFFF);  // White
  end;

  //Draw endpoints in legend chart so that colours show up
  With Chart2LineSeries1 do begin
    Clear;
    AddXY(0.0,LegendMin);
    AddXY(0.0,LegendMax);
  end;

  Chart2.AxisList[0].Range.Min:=LegendMin;
  Chart2.AxisList[0].Range.Max:=LegendMax;

end;

procedure TDLRetrieveForm.FormShow(Sender: TObject);
begin

  //Resize for small screens
  if DLRetrieveForm.Width>Screen.Width then DLRetrieveForm.Width:=Screen.Width;
  if DLRetrieveForm.Height>Screen.Height then DLRetrieveForm.Height:=Screen.Height;
  if DLRetrieveForm.Left<0 then DLRetrieveForm.Left:=0;
  if DLRetrieveForm.Top<0 then DLRetrieveForm.Top:=0;

  //FileDirectoryEdit.Text:=Format('%s%',[appsettings.LogsDirectory + DirectorySeparator]);
  LogsDirectoryDisplay.Text:= RemoveMultiSlash(Format('%s%',[appsettings.LogsDirectory + DirectorySeparator]));

   //Determine if vector plot can be shown
   if (SelectedModel<>model_V) and (not VectorPlotOverride) then
     PageControl1.Pages[1].TabVisible:=False;

   //Show how many records can be retreived
   MaxRecordsLabel.Caption:='['+IntToStr(DLEStoredRecords)+' max]';
end;

procedure TDLRetrieveForm.OpenAnotherFileButtonClick(Sender: TObject);
var
  Filename: String;
begin
  OpenDialog1.InitialDir:=appsettings.LogsDirectory;
  OpenDialog1.Execute;

  Filename:=OpenDialog1.FileName;
  RecentFileEdit.Text:=Filename;

  if FileExists(Filename) then
    SynMemo2.Lines.LoadFromFile(Filename)
    else
       begin
        SynMemo2.Clear;
        SynMemo2.Lines.Append('File:');
        SynMemo2.Lines.Append(format('    %s',[Filename]));
        SynMemo2.Lines.Append('does not exist!');
      end;
end;

procedure TDLRetrieveForm.OpenDirectoryButtonClick(Sender: TObject);
begin
  OpenDialog1.InitialDir := appsettings.LogsDirectory;
  if OpenDialog1.Execute then
  begin{ TODO : file viewer form2, make something new maybe. }
    SynMemo2.Lines.LoadFromFile(OpenDialog1.Filename);
    //Form2.Memo1.Lines.LoadFromFile(OpenDialog1.Filename);
    //Form2.Show;
  end;

end;

procedure TDLRetrieveForm.OpenRecentFileButtonClick(Sender: TObject);
var
  RecentFilname: String;
begin

  RecentFilname:=RecentFileEdit.Text;

  if FileExists(RecentFilname) then
    SynMemo2.Lines.LoadFromFile(RecentFilname)
    else
       begin
        SynMemo2.Clear;
        SynMemo2.Lines.Append('File:');
        SynMemo2.Lines.Append(format('    %s',[RecentFilname]));
        SynMemo2.Lines.Append('does not exist!');
      end;
end;

procedure TDLRetrieveForm.ShowDotsCheckBoxChange(Sender: TObject);
begin
  //Show hourglass cursor while changing display.
  HourGlass();
  DotsLinesUpdate();
end;

procedure TDLRetrieveForm.ShowLinesCheckBoxChange(Sender: TObject);
begin
  //Show hourglass cursor while changing display.
  HourGlass();
  DotsLinesUpdate();
end;

procedure TDLRetrieveForm.DotsLinesUpdate(); begin

  //Dots only = black ring with red dot in center for higher visibility

  if not ShowDotsCheckBox.Checked and not ShowLinesCheckBox.Checked then begin
    Chart1LineSeries1.Pointer.Visible := False;
    Chart1LineSeries2.LinePen.Style:=psClear;
    Chart1LineSeries2.LinePen.Mode:=pmNop;
  end;

  if not ShowDotsCheckBox.Checked and ShowLinesCheckBox.Checked then begin
    Chart1LineSeries1.Pointer.Visible := False;
    Chart1LineSeries2.LinePen.Style:=psSolid;
    Chart1LineSeries2.LinePen.Mode:=pmCopy;
  end;

  if ShowDotsCheckBox.Checked and not ShowLinesCheckBox.Checked then begin
    Chart1LineSeries1.Pointer.Visible := True;
    Chart1LineSeries2.LinePen.Style:=psClear;
    Chart1LineSeries2.LinePen.Mode:=pmCopy;
  end;

  if ShowDotsCheckBox.Checked and ShowLinesCheckBox.Checked then begin
    Chart1LineSeries1.Pointer.Visible := True;
    Chart1LineSeries2.LinePen.Style:=psSolid;
    Chart1LineSeries2.LinePen.Mode:=pmCopy;
  end;

end;

procedure TDLRetrieveForm.HourGlass();
begin
  //Show hourglass cursor.
  HourGlassTimeout:=0;
  CalculatingText.Visible:=True;
  CalculatingProgressBar.Visible:=True;
end;

initialization
  {$I dlretrieve.lrs}

Finalization

end.

